@@ -10,10 +10,11 @@ __all__ = ['solve', 'solve_triangular', 'solveh_banded', 'solve_banded',
            'solve_toeplitz', 'solve_circulant', 'inv', 'det', 'lstsq',
            'pinv', 'pinv2', 'pinvh']
 
+import warnings
 import numpy as np
 
 from .flinalg import get_flinalg_funcs
-from .lapack import get_lapack_funcs
+from .lapack import get_lapack_funcs, _compute_lwork
 from .misc import LinAlgError, _datacopied
 from .decomp import _asarray_validated
 from . import decomp, decomp_svd
@@ -58,6 +59,8 @@ def solve(a, b, sym_pos=False, lower=False, overwrite_a=False,
     ------
     LinAlgError
         If `a` is singular.
+    ValueError
+        If `a` is not square
 
     Examples
     --------
@@ -82,24 +85,24 @@ def solve(a, b, sym_pos=False, lower=False, overwrite_a=False,
     overwrite_a = overwrite_a or _datacopied(a1, a)
     overwrite_b = overwrite_b or _datacopied(b1, b)
     if debug:
-        print('solve:overwrite_a=',overwrite_a)
-        print('solve:overwrite_b=',overwrite_b)
+        print('solve:overwrite_a=', overwrite_a)
+        print('solve:overwrite_b=', overwrite_b)
     if sym_pos:
-        posv, = get_lapack_funcs(('posv',), (a1,b1))
+        posv, = get_lapack_funcs(('posv',), (a1, b1))
         c, x, info = posv(a1, b1, lower=lower,
-                        overwrite_a=overwrite_a,
-                        overwrite_b=overwrite_b)
+                          overwrite_a=overwrite_a,
+                          overwrite_b=overwrite_b)
     else:
-        gesv, = get_lapack_funcs(('gesv',), (a1,b1))
+        gesv, = get_lapack_funcs(('gesv',), (a1, b1))
         lu, piv, x, info = gesv(a1, b1, overwrite_a=overwrite_a,
-                                            overwrite_b=overwrite_b)
+                                overwrite_b=overwrite_b)
 
     if info == 0:
         return x
     if info > 0:
         raise LinAlgError("singular matrix")
-    raise ValueError('illegal value in %d-th argument of internal gesv|posv'
-                                                                    % -info)
+    raise ValueError('illegal value in %d-th argument of internal gesv|posv' %
+                     -info)
 
 
 def solve_triangular(a, b, trans=0, lower=False, unit_diagonal=False,
@@ -159,22 +162,23 @@ def solve_triangular(a, b, trans=0, lower=False, unit_diagonal=False,
         raise ValueError('incompatible dimensions')
     overwrite_b = overwrite_b or _datacopied(b1, b)
     if debug:
-        print('solve:overwrite_b=',overwrite_b)
+        print('solve:overwrite_b=', overwrite_b)
     trans = {'N': 0, 'T': 1, 'C': 2}.get(trans, trans)
-    trtrs, = get_lapack_funcs(('trtrs',), (a1,b1))
+    trtrs, = get_lapack_funcs(('trtrs',), (a1, b1))
     x, info = trtrs(a1, b1, overwrite_b=overwrite_b, lower=lower,
                     trans=trans, unitdiag=unit_diagonal)
 
     if info == 0:
         return x
     if info > 0:
-        raise LinAlgError("singular matrix: resolution failed at diagonal %s" % (info-1))
-    raise ValueError('illegal value in %d-th argument of internal trtrs'
-            % -info)
+        raise LinAlgError("singular matrix: resolution failed at diagonal %d" %
+                          (info-1))
+    raise ValueError('illegal value in %d-th argument of internal trtrs' %
+                     (-info))
 
 
 def solve_banded(l_and_u, ab, b, overwrite_ab=False, overwrite_b=False,
-                debug=False, check_finite=True):
+                 debug=False, check_finite=True):
     """
     Solve the equation a x = b for x, assuming a is banded matrix.
 
@@ -220,37 +224,39 @@ def solve_banded(l_and_u, ab, b, overwrite_ab=False, overwrite_b=False,
         raise ValueError("shapes of ab and b are not compatible.")
     (l, u) = l_and_u
     if l + u + 1 != a1.shape[0]:
-        raise ValueError("invalid values for the number of lower and upper diagonals:"
-                " l+u+1 (%d) does not equal ab.shape[0] (%d)" % (l+u+1, ab.shape[0]))
+        raise ValueError("invalid values for the number of lower and upper "
+                         "diagonals: l+u+1 (%d) does not equal ab.shape[0] "
+                         "(%d)" % (l+u+1, ab.shape[0]))
 
     overwrite_b = overwrite_b or _datacopied(b1, b)
     if a1.shape[-1] == 1:
         b2 = np.array(b1, copy=overwrite_b)
-        b2 /= a1[1,0]
+        b2 /= a1[1, 0]
         return b2
     if l == u == 1:
         overwrite_ab = overwrite_ab or _datacopied(a1, ab)
         gtsv, = get_lapack_funcs(('gtsv',), (a1, b1))
-        du = a1[0,1:]
-        d = a1[1,:]
-        dl = a1[2,:-1]
+        du = a1[0, 1:]
+        d = a1[1, :]
+        dl = a1[2, :-1]
         du2, d, du, x, info = gtsv(dl, d, du, b1, overwrite_ab, overwrite_ab,
                                    overwrite_ab, overwrite_b)
     else:
         gbsv, = get_lapack_funcs(('gbsv',), (a1, b1))
         a2 = np.zeros((2*l+u+1, a1.shape[1]), dtype=gbsv.dtype)
-        a2[l:,:] = a1
+        a2[l:, :] = a1
         lu, piv, x, info = gbsv(l, u, a2, b1, overwrite_ab=True,
-                                                    overwrite_b=overwrite_b)
+                                overwrite_b=overwrite_b)
     if info == 0:
         return x
     if info > 0:
         raise LinAlgError("singular matrix")
-    raise ValueError('illegal value in %d-th argument of internal gbsv/gtsv' % -info)
+    raise ValueError('illegal value in %d-th argument of internal gbsv/gtsv' %
+                     -info)
 
 
 def solveh_banded(ab, b, overwrite_ab=False, overwrite_b=False, lower=False,
-                    check_finite=True):
+                  check_finite=True):
     """
     Solve equation a x = b. a is Hermitian positive-definite banded matrix.
 
@@ -310,21 +316,22 @@ def solveh_banded(ab, b, overwrite_ab=False, overwrite_b=False, lower=False,
     if a1.shape[0] == 2:
         ptsv, = get_lapack_funcs(('ptsv',), (a1, b1))
         if lower:
-            d = a1[0,:].real
-            e = a1[1,:-1]
+            d = a1[0, :].real
+            e = a1[1, :-1]
         else:
-            d = a1[1,:].real
-            e = a1[0,1:].conj()
-        d, du, x, info = ptsv(d, e, b1, overwrite_ab, overwrite_ab, overwrite_b)
+            d = a1[1, :].real
+            e = a1[0, 1:].conj()
+        d, du, x, info = ptsv(d, e, b1, overwrite_ab, overwrite_ab,
+                              overwrite_b)
     else:
         pbsv, = get_lapack_funcs(('pbsv',), (a1, b1))
         c, x, info = pbsv(a1, b1, lower=lower, overwrite_ab=overwrite_ab,
-                                                overwrite_b=overwrite_b)
+                          overwrite_b=overwrite_b)
     if info > 0:
         raise LinAlgError("%d-th leading minor not positive definite" % info)
     if info < 0:
-        raise ValueError('illegal value in %d-th argument of internal pbsv'
-                                                                    % -info)
+        raise ValueError('illegal value in %d-th argument of internal pbsv' %
+                         -info)
     return x
 
 
@@ -362,7 +369,7 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
     The solution is computed using Levinson-Durbin recursion, which is faster
     than generic least-squares methods, but can be less numerically stable.
     """
-    # If numerical stability of this algorithim is a problem, a future
+    # If numerical stability of this algorithm is a problem, a future
     # developer might consider implementing other O(N^2) Toeplitz solvers,
     # such as GKO (http://www.jstor.org/stable/2153371) or Bareiss.
     if isinstance(c_or_cr, tuple):
@@ -396,7 +403,7 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
         b_shape = b.shape
         b = b.reshape(b.shape[0], -1)
         x = np.column_stack(
-            (levinson(vals, np.ascontiguousarray(b[:,i]))[0])
+            (levinson(vals, np.ascontiguousarray(b[:, i]))[0])
             for i in range(b.shape[1]))
         x = x.reshape(*b_shape)
 
@@ -631,9 +638,9 @@ def inv(a, overwrite_a=False, check_finite=True):
 
     Raises
     ------
-    LinAlgError :
+    LinAlgError
         If `a` is singular.
-    ValueError :
+    ValueError
         If `a` is not square, or not 2-dimensional.
 
     Examples
@@ -661,13 +668,12 @@ def inv(a, overwrite_a=False, check_finite=True):
 ##         if info>0: raise LinAlgError, "singular matrix"
 ##         if info<0: raise ValueError,\
 ##            'illegal value in %d-th argument of internal inv.getrf|getri'%(-info)
-    getrf, getri, getri_lwork = get_lapack_funcs(('getrf','getri', 'getri_lwork'), (a1,))
+    getrf, getri, getri_lwork = get_lapack_funcs(('getrf', 'getri',
+                                                  'getri_lwork'),
+                                                 (a1,))
     lu, piv, info = getrf(a1, overwrite_a=overwrite_a)
     if info == 0:
-        lwork, info = getri_lwork(a1.shape[0])
-        if info != 0:
-            raise ValueError('internal getri work space query failed: %d' % (info,))
-        lwork = int(lwork.real)
+        lwork = _compute_lwork(getri_lwork, a1.shape[0])
 
         # XXX: the following line fixes curious SEGFAULT when
         # benchmarking 500x500 matrix inverse. This seems to
@@ -681,7 +687,7 @@ def inv(a, overwrite_a=False, check_finite=True):
         raise LinAlgError("singular matrix")
     if info < 0:
         raise ValueError('illegal value in %d-th argument of internal '
-                                                    'getrf|getri' % -info)
+                         'getrf|getri' % -info)
     return inv_a
 
 
@@ -741,14 +747,17 @@ def det(a, overwrite_a=False, check_finite=True):
     a_det, info = fdet(a1, overwrite_a=overwrite_a)
     if info < 0:
         raise ValueError('illegal value in %d-th argument of internal '
-                                                        'det.getrf' % -info)
+                         'det.getrf' % -info)
     return a_det
 
 ### Linear Least Squares
 
+class LstsqLapackError(LinAlgError):
+    pass
+
 
 def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
-          check_finite=True):
+          check_finite=True, lapack_driver=None):
     """
     Compute least-squares solution to equation Ax = b.
 
@@ -772,6 +781,14 @@ def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
         Whether to check that the input matrices contain only finite numbers.
         Disabling may give a performance gain, but may result in problems
         (crashes, non-termination) if the inputs do contain infinities or NaNs.
+    lapack_driver: str, optional
+        Which LAPACK driver is used to solve the least-squares problem.
+        Options are ``'gelsd'``, ``'gelsy'``, ``'gelss'``. Default
+        (``'gelsd'``) is a good choice.  However, ``'gelsy'`` can be slightly
+        faster on many problems.  ``'gelss'`` was used historically.  It is
+        generally slow but uses less memory.
+
+        .. versionadded:: 0.17.0
 
     Returns
     -------
@@ -779,19 +796,22 @@ def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
         Least-squares solution.  Return shape matches shape of `b`.
     residues : () or (1,) or (K,) ndarray
         Sums of residues, squared 2-norm for each column in ``b - a x``.
-        If rank of matrix a is < N or > M this is an empty array.
-        If b was 1-D, this is an (1,) shape array, otherwise the shape is (K,).
+        If rank of matrix a is ``< N`` or ``> M``, or ``'gelsy'`` is used,
+        this is an empty array. If b was 1-D, this is an (1,) shape array,
+        otherwise the shape is (K,).
     rank : int
         Effective rank of matrix `a`.
-    s : (min(M,N),) ndarray
+    s : (min(M,N),) ndarray or None
         Singular values of `a`. The condition number of a is
-        ``abs(s[0]/s[-1])``.
+        ``abs(s[0] / s[-1])``. None is returned when ``'gelsy'`` is used.
 
     Raises
     ------
-    LinAlgError :
+    LinAlgError
         If computation does not converge.
 
+    ValueError
+        When parameters are wrong.
 
     See Also
     --------
@@ -809,41 +829,100 @@ def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
         nrhs = 1
     if m != b1.shape[0]:
         raise ValueError('incompatible dimensions')
-    gelss, = get_lapack_funcs(('gelss',), (a1, b1))
-    if n > m:
+
+    driver = lapack_driver
+    if driver is None:
+        driver = lstsq.default_lapack_driver
+    if driver not in ('gelsd', 'gelsy', 'gelss'):
+        raise ValueError('LAPACK driver "%s" is not found' % driver)
+
+    lapack_func, lapack_lwork = get_lapack_funcs((driver,
+                                        '%s_lwork' % driver), (a1, b1))
+    real_data = True if (lapack_func.dtype.kind == 'f') else False
+
+    if m < n:
         # need to extend b matrix as it will be filled with
         # a larger solution matrix
         if len(b1.shape) == 2:
-            b2 = np.zeros((n, nrhs), dtype=gelss.dtype)
-            b2[:m,:] = b1
+            b2 = np.zeros((n, nrhs), dtype=lapack_func.dtype)
+            b2[:m, :] = b1
         else:
-            b2 = np.zeros(n, dtype=gelss.dtype)
+            b2 = np.zeros(n, dtype=lapack_func.dtype)
             b2[:m] = b1
         b1 = b2
 
     overwrite_a = overwrite_a or _datacopied(a1, a)
     overwrite_b = overwrite_b or _datacopied(b1, b)
 
-    # get optimal work array
-    work = gelss(a1, b1, lwork=-1)[4]
-    lwork = work[0].real.astype(np.int)
-    v, x, s, rank, work, info = gelss(
-        a1, b1, cond=cond, lwork=lwork, overwrite_a=overwrite_a,
-        overwrite_b=overwrite_b)
+    if cond is None:
+        cond = np.finfo(lapack_func.dtype).eps
 
-    if info > 0:
-        raise LinAlgError("SVD did not converge in Linear Least Squares")
-    if info < 0:
-        raise ValueError('illegal value in %d-th argument of internal gelss'
-                                                                    % -info)
-    resids = np.asarray([], dtype=x.dtype)
-    if n < m:
-        x1 = x[:n]
-        if rank == n:
-            resids = np.sum(np.abs(x[n:])**2, axis=0)
-        x = x1
-    return x, resids, rank, s
+    if driver in ('gelss', 'gelsd'):
+        if driver == 'gelss':
+            lwork = _compute_lwork(lapack_lwork, m, n, nrhs, cond)
+            v, x, s, rank, work, info = lapack_func(a1, b1, cond, lwork,
+                                                    overwrite_a=overwrite_a,
+                                                    overwrite_b=overwrite_b)
 
+        elif driver == 'gelsd':
+            if real_data:
+                lwork, iwork = _compute_lwork(lapack_lwork, m, n, nrhs, cond)
+                if iwork == 0:
+                    # this is LAPACK bug 0038: dgelsd does not provide the
+                    # size of the iwork array in query mode.  This bug was
+                    # fixed in LAPACK 3.2.2, released July 21, 2010.
+                    mesg = ("internal gelsd driver lwork query error, "
+                           "required iwork dimension not returned. "
+                           "This is likely the result of LAPACK bug "
+                           "0038, fixed in LAPACK 3.2.2 (released "
+                           "July 21, 2010). ")
+                    
+                    if lapack_driver is None:
+                        # restart with gelss
+                        lstsq.default_lapack_driver = 'gelss'
+                        mesg += "Falling back to 'gelss' driver."
+                        warnings.warn(mesg, RuntimeWarning)
+                        return lstsq(a, b, cond, overwrite_a, overwrite_b,
+                                    check_finite, lapack_driver='gelss')
+
+                    # can't proceed, bail out
+                    mesg += ("Use a different lapack_driver when calling lstsq "
+                            "or upgrade LAPACK.")
+                    raise LstsqLapackError(mesg)
+
+                x, s, rank, info = lapack_func(a1, b1, lwork,
+                                               iwork, cond, False, False)
+            else:  # complex data
+                lwork, rwork, iwork = _compute_lwork(lapack_lwork, m, n,
+                                                     nrhs, cond)
+                x, s, rank, info = lapack_func(a1, b1, lwork, rwork, iwork,
+                                               cond, False, False)
+        if info > 0:
+            raise LinAlgError("SVD did not converge in Linear Least Squares")
+        if info < 0:
+            raise ValueError('illegal value in %d-th argument of internal %s'
+                                                    % (-info, lapack_driver))
+        resids = np.asarray([], dtype=x.dtype)
+        if m > n:
+            x1 = x[:n]
+            if rank == n:
+                resids = np.sum(np.abs(x[n:])**2, axis=0)
+            x = x1
+        return x, resids, rank, s
+
+    elif driver == 'gelsy':
+        lwork = _compute_lwork(lapack_lwork, m, n, nrhs, cond)
+        jptv = np.zeros((a1.shape[1],1), dtype=np.int32)
+        v, x, j, rank, info = lapack_func(a1, b1, jptv, cond,
+                                           lwork, False, False)
+        if info < 0:
+            raise ValueError("illegal value in %d-th argument of internal "
+                             "gelsy" % -info)
+        if m > n:
+            x1 = x[:n]
+            x = x1
+        return x, np.array([], x.dtype), rank, None
+lstsq.default_lapack_driver = 'gelsd'
 
 def pinv(a, cond=None, rcond=None, return_rank=False, check_finite=True):
     """
@@ -955,16 +1034,16 @@ def pinv2(a, cond=None, rcond=None, return_rank=False, check_finite=True):
 
     if rcond is not None:
         cond = rcond
-    if cond in [None,-1]:
+    if cond in [None, -1]:
         t = u.dtype.char.lower()
         factor = {'f': 1E3, 'd': 1E6}
         cond = factor[t] * np.finfo(t).eps
 
     rank = np.sum(s > cond * np.max(s))
-    psigma_diag = 1.0 / s[: rank]
 
-    B = np.transpose(np.conjugate(np.dot(u[:, : rank] *
-                                         psigma_diag, vh[: rank])))
+    u = u[:, :rank]
+    u /= s[:rank]
+    B = np.transpose(np.conjugate(np.dot(u, vh[:rank])))
 
     if return_rank:
         return B, rank
